use crate::types;
use serde::{
    de::{self, Deserializer},
    ser::{self, SerializeSeq, Serializer},
    Deserialize, Serialize,
};

#[derive(Debug, Clone, Default)]
pub struct Request {
    pub subscription_id: types::SubscriptionId,
    pub filters: Vec<types::Filter>,
}

impl<'de> de::Deserialize<'de> for Request {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        #[derive(Deserialize, Serialize, Debug)]
        #[serde(untagged)]
        enum StringOrFilter {
            Filter(types::Filter),
            String(String),
        }

        impl<'de> StringOrFilter {
            pub fn to_filter<D>(&self, element: usize) -> Result<types::Filter, D::Error>
            where
                D: Deserializer<'de>,
            {
                match self {
                    Self::Filter(f) => Ok(f.clone()),
                    _ => Err(de::Error::custom(format!(
                        "Expecting a filter in element {}, got a filter object",
                        element,
                    ))),
                }
            }

            pub fn to_string<D>(&self, element: usize) -> Result<String, D::Error>
            where
                D: Deserializer<'de>,
            {
                match self {
                    Self::String(string) => Ok(string.to_owned()),
                    _ => Err(de::Error::custom(format!(
                        "Expecting an string in element {}, got a filter object",
                        element,
                    ))),
                }
            }
        }

        let s: Vec<StringOrFilter> = Deserialize::deserialize(deserializer)?;
        if s.len() < 2 {
            return Err(de::Error::custom(
                "Array too small, it must have at least 2 elements",
            ));
        }
        let header = s[0].to_string::<D>(0)?;
        let subscription_id = s[1]
            .to_string::<D>(1)?
            .as_str()
            .try_into()
            .map_err(|e: types::subscription_id::Error| de::Error::custom(e.to_string()))?;

        if header != "REQ" {
            return Err(de::Error::custom(format!(
                "Invalid header, got {} and expected REQ",
                header
            )));
        }

        let mut index = 1;
        Ok(Request {
            subscription_id,
            filters: if s.len() > 2 {
                s[2..]
                    .iter()
                    .map(|filter| {
                        index += 1;
                        filter.to_filter::<D>(index)
                    })
                    .collect::<Result<Vec<types::Filter>, D::Error>>()?
            } else {
                vec![]
            },
        })
    }
}

impl ser::Serialize for Request {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut seq = serializer.serialize_seq(Some(self.filters.len() + 2))?;
        seq.serialize_element("REQ")?;
        seq.serialize_element(&self.subscription_id)?;
        for element in self.filters.iter() {
            seq.serialize_element(element)?;
        }
        seq.end()
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use chrono::Utc;

    #[test]
    fn encoding_decoding() {
        let r = Request {
            filters: vec![
                types::Filter {
                    authors: vec![
                        "npub1h0h2zfcp3z7acvc9l0qvpgfqznkalj77whsdkmmlwanfssys3ngsx336a8"
                            .try_into()
                            .expect("some addr object"),
                    ],
                    until: Some(Utc::now()),
                    limit: 30,
                    ..Default::default()
                },
                types::Filter {
                    limit: 20,
                    ..Default::default()
                },
            ],
            ..Default::default()
        };

        let serialized = serde_json::to_string(&r).expect("valid json string");
        let obj: Request = serde_json::from_str(&serialized).expect("valid request");
        assert_eq!(r.subscription_id, obj.subscription_id);
        assert_eq!(r.filters.len(), obj.filters.len());
    }
}
