use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Timestamp {0} is ambiguous (min={1}, max={2})")]
    AmbiguousTimestamp(i64, DateTime<Utc>, DateTime<Utc>),
}

#[derive(Serialize, Deserialize, Default, Debug, Clone)]
pub struct Filter {
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub ids: Vec<super::Addr>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub authors: Vec<super::Addr>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub kinds: Vec<u32>,
    #[serde(default, skip_serializing_if = "Vec::is_empty")]
    pub tags: Vec<super::Tag>,
    #[serde(default, with = "ts_seconds", skip_serializing_if = "Option::is_none")]
    pub since: Option<DateTime<Utc>>,
    #[serde(default, with = "ts_seconds", skip_serializing_if = "Option::is_none")]
    pub until: Option<DateTime<Utc>>,
    #[serde(default, skip_serializing_if = "is_zero")]
    pub limit: u64,
}

mod ts_seconds {
    use super::Error;
    use chrono::{DateTime, LocalResult, TimeZone, Utc};
    use serde::{de, Deserialize, Deserializer, Serializer};

    pub fn serialize<S>(date: &Option<DateTime<Utc>>, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        match date {
            Some(date) => serializer.serialize_i64(date.timestamp()),
            _ => serializer.serialize_none(),
        }
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Option<DateTime<Utc>>, D::Error>
    where
        D: Deserializer<'de>,
    {
        let timestamp: Option<i64> = Deserialize::deserialize(deserializer)?;
        let timestamp = if let Some(timestamp) = timestamp {
            timestamp
        } else {
            return Ok(None);
        };

        match Utc.timestamp_opt(timestamp, 0) {
            LocalResult::None => Err(de::Error::custom("Invalid timestamp")),
            LocalResult::Ambiguous(min, max) => Err(de::Error::custom(
                Error::AmbiguousTimestamp(timestamp, min, max).to_string(),
            )),
            LocalResult::Single(ts) => Ok(Some(ts)),
        }
    }
}

#[inline]
fn is_zero(number: &u64) -> bool {
    *number == 0
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn deserialize() {
        let json = "{\"authors\":[\"c0f0f1\", \"1234567901\"]}";
        let obj: Result<Filter, _> = serde_json::from_str(&json);
        assert!(obj.is_ok());
    }

    #[test]
    fn serialize() {
        let now = Utc::now();
        let x = Filter {
            since: Some(now),
            ..Default::default()
        };

        assert_eq!(
            format!("{}\"since\":{:?}{}", "{", now.timestamp(), "}"),
            serde_json::to_string(&x).expect("valid json")
        );
    }
}
