use hex::ToHex;
use rand::RngCore;
use serde::{
    de::{self, Deserialize, Deserializer},
    Serialize,
};
use std::convert::TryFrom;
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Subscription ID is limited to 64 characters")]
    TooLong,
}

#[derive(Serialize, Debug, Clone, PartialEq, Eq)]
pub struct SubscriptionId(String);

impl<'de> Deserialize<'de> for SubscriptionId {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s = <&str>::deserialize(deserializer)?;
        let t = s
            .try_into()
            .map_err(|e: Error| de::Error::custom(e.to_string()))?;
        Ok(t)
    }
}

impl TryFrom<&str> for SubscriptionId {
    type Error = Error;
    fn try_from(s: &str) -> Result<Self, Self::Error> {
        if s.as_bytes().len() > 64 {
            return Err(Error::TooLong);
        }
        Ok(SubscriptionId(s.into()))
    }
}

impl Default for SubscriptionId {
    fn default() -> Self {
        let mut data = [0u8; 32];
        rand::thread_rng().fill_bytes(&mut data);
        Self(data.encode_hex::<String>())
    }
}

impl ToString for SubscriptionId {
    fn to_string(&self) -> String {
        self.0.clone()
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn default() {
        assert_eq!(64, SubscriptionId::default().to_string().len());
    }

    #[test]
    fn too_long() {
        let f = format!("too_long:{}", SubscriptionId::default().to_string());
        let r: Result<SubscriptionId, _> = f.as_str().try_into();
        assert!(r.is_err());
    }

    #[test]
    fn deserialize() {
        let r: Result<SubscriptionId, _> = serde_json::from_str("\"ok\"");
        assert_eq!("ok".to_owned(), r.expect("valid").to_string());
    }

    #[test]
    fn deserialize_err() {
        let f = format!("\"too_long:{}\"", SubscriptionId::default().to_string());
        let r: Result<SubscriptionId, _> = serde_json::from_str(&f);
        assert!(r.is_err());
    }
}
